# Copyright (C) 2007, 2008 by Intevation GmbH
# Authors:
# Bernhard Herzog <bh@intevation.de>
#
# This program is free software under the GPL (>=v2)
# Read the file COPYING coming with the software for details.

"""GetCapabilities Fees/AccessConstraints processing"""

import urlparse
from StringIO import StringIO

from lxml import etree

import inteproxy.main


# keep track of the URLs for which the dialog was already shown.
dialog_shown_for = dict()


def is_get_capabilities_url(url):
    scheme, netloc, path, query, fragment = urlparse.urlsplit(url)

    if query:
        split_query = query.split("&")
        for params in split_query:
            parts = params.split("=")
            if parts[0].upper() == "REQUEST" and parts[1] == "GetCapabilities":
                return True

    return False


def parse_capabilities(xml):
    """Parses the capabilities information and returns some of the information

    The return value is a tuple (title, fees, access_constraints) with
    the contents of the correspondings sub-elements of the xml
    structures Service element"""
    t = etree.parse(StringIO(xml))
    title = t.findtext("Service/Title")
    fees = t.findtext("Service/Fees")
    access_constraints = t.findtext("Service/AccessConstraints")
    return (title, fees, access_constraints)



def handle_fees_and_access_constraints(remote_url, response):
    """Runs Fees/AccessConstraints dialog for GetCapabilities responses

    This function uses is_get_capabilities_url to determine whether the
    remote_url was a GetCapabilities request.  If that is the case, the
    response is assumed to be a GetCapabilities response.  This function
    then parses the response as XML and extract Fees and
    AccessConstraints information from it and shows them in a dialog.

    The response parameter should be the HTTPRequestMessage object with
    the response from the remote url.

    The function keeps track of which URLs were already processed, so
    that the dialog is never shown twice for the same URL.
    """
    # only try to handle anything for successful requests.
    # Should we allow any other success response codes?
    if response.status != 200:
        return

    if remote_url in dialog_shown_for:
        return

    if is_get_capabilities_url(remote_url):
        title, fees, access_constraints = parse_capabilities(response.body)
        inteproxy.main.the_application.run_fees_dialog(title, fees,
                                                       access_constraints)

        dialog_shown_for[remote_url] = True
