# Copyright (C) 2007, 2009 by Intevation GmbH
# Authors:
# Bernhard Herzog <bh@intevation.de>
#
# This program is free software under the GPL (>=v2)
# Read the file COPYING coming with the software for details.

"""HTTP server that can run in a thread and be stopped"""

import socket
import errno
import threading
import BaseHTTPServer


class HTTPServer(BaseHTTPServer.HTTPServer):

    """HTTP server that can be run in a thread and be stopped"""

    do_shutdown = False

    def getsockname(self):
        """Return the socket's sockname as a (host, port) tuple"""
        return self.socket.getsockname()

    def serve_forever(self):
        """Handle requests until self.do_shutdown is True."""
        while not self.do_shutdown:
            self.handle_request()

    def shutdown(self):
        """Shutdown the server.

        Usually this is called from a thread while another thread
        executes the serve_forever method.  In that case the other
        thread is most likely blocked while listening for new
        connections.  After this method returns, the caller will have to
        do something to unblock the other thread, such as creating a TCP
        connection to the server.
        """
        self.do_shutdown = True


class ServerThread(object):

    """Class to run a HTTPServer instance in a thread"""

    def __init__(self, server):
        """Initializes the server thread with an instance of HTTPServer"""
        self.server = server
        self.server_port = self.server.getsockname()[1]
        self.server_thread = None

    def start(self, daemon=False):
        """Starts the server thread"""
        self.server_thread = threading.Thread(target=self._serve_forever)
        self.server_thread.setDaemon(daemon)
        self.server_thread.start()

    def _serve_forever(self):
        """Helper method to run the server's serve_forever method in a thread"""
        self.server.serve_forever()

    def stop(self):
        """Stops the server thread"""
        if self.server_thread is None:
            return

        self.server.shutdown()

        # The server thread might be blocked while listening on the
        # socket.  Unblock it by connecting to the port.
        s = None
        try:
            s = socket.socket()
            s.connect(("localhost", self.server_port))
            s.shutdown(socket.SHUT_WR)
        except socket.error, exc:
            # The server may have shut down already when we try to
            # connect, so we ignore connection failures.
            if exc[0] == errno.ECONNREFUSED:
                pass
            else:
                raise

        self.server_thread.join()
        self.server_thread = None

        self.server.server_close()

        if s is not None:
            s.close()
